/*****************************************************************************

	emumain.c

	߫죭󫳫

******************************************************************************/

#include "emumain.h"
#include "inptport.h"
#include <stdarg.h>
#ifndef GP2X
#include <SDL.h>
#endif

#define FRAMESKIP_LEVELS	12

#ifdef GP2X
#define TICKS_PER_SEC 1000000UL
#else
#define TICKS_PER_SEC 1000UL
#endif
#define TICKS_PER_FRAME		(TICKS_PER_SEC / FPS)


/******************************************************************************
	Ыܨ
******************************************************************************/

u8 resize_table[320];

char game_name[16] = "ddsom";
char parent_name[16];
char cache_parent_name[16];

char game_dir[MAX_PATH] = "./roms";
char cache_dir[MAX_PATH] = "./cache";

int option_showfps;
int option_speedlimit;
int option_autoframeskip;
int option_frameskip;
int option_vsync;
int option_stretch = 0;

int option_sound_enable;
int option_samplerate;
int option_sound_volume;
int option_sound_filter;

int machine_driver_type;
int machine_init_type;
int machine_input_type;
int machine_screen_type;
int machine_sound_type;

u32 frames_displayed;
int fatal_error;
static int msg_line = 0;
static int msg_count = 0;

#ifdef GP2X
int mem_fd = -1;
int joy_fd = -1;
u8 *upper_memory = NULL;
#endif

volatile int Loop;
volatile int Sleep;
char launchDir[MAX_PATH];


/******************************************************************************
	ܨ
******************************************************************************/

static int frameskip;
static int frameskipadjust;
static int frameskip_counter;

static TICKER next_frame_base;
//static TICKER last_skipcount0_time;
//static int warming_up;

static int frames_since_last_fps;
static int rendered_frames_since_last_fps;
static int game_speed_percent;
static int frames_per_second;
static int skip_nextframe;

static int snap_no = -1;

static char fatal_error_message[256];

static const u8 skiptable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
{
	{ 0,0,0,0,0,0,0,0,0,0,0,0 },
	{ 0,0,0,0,0,0,0,0,0,0,0,1 },
	{ 0,0,0,0,0,1,0,0,0,0,0,1 },
	{ 0,0,0,1,0,0,0,1,0,0,0,1 },
	{ 0,0,1,0,0,1,0,0,1,0,0,1 },
	{ 0,1,0,0,1,0,1,0,0,1,0,1 },
	{ 0,1,0,1,0,1,0,1,0,1,0,1 },
	{ 0,1,0,1,1,0,1,0,1,1,0,1 },
	{ 0,1,1,0,1,1,0,1,1,0,1,1 },
	{ 0,1,1,1,0,1,1,1,0,1,1,1 },
	{ 0,1,1,1,1,1,0,1,1,1,1,1 },
	{ 0,1,1,1,1,1,1,1,1,1,1,1 }
};


/******************************************************************************
	׫ȫ
******************************************************************************/

static void show_fps(void);
static void show_battery_warning(void);
static void draw_text(const char *text);

/*--------------------------------------------------------
	߫죭
--------------------------------------------------------*/

#ifndef GP2X
SDL_Surface *screen_surface = NULL;
SDL_Surface *work_surface = NULL;
#endif
u16 *work_frame = NULL;
static u16 *screen = NULL;

void Terminate(void)
{
#ifdef GP2X
    if(work_frame) free(work_frame);
    if(mem_fd >= 0) close(mem_fd);
    if(joy_fd >= 0) close(joy_fd);
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#else
    if(work_surface)
        SDL_FreeSurface(work_surface);
    SDL_Quit();
#endif
}

int main(int argc, char *argv[])
{
#ifdef GP2X
/*
	if (SDL_Init (SDL_INIT_AUDIO) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit (1);
	}
*/
	atexit (Terminate);

    joy_fd = open("/dev/GPIO", O_RDWR | O_NDELAY);
    if(joy_fd < 0) {
		fprintf (stderr, "Couldn't open /dev/GPIO device.\n");
		exit (1);
    }

    mem_fd = open("/dev/mem", O_RDWR | O_SYNC);
    if(mem_fd < 0) {
		fprintf (stderr, "Couldn't open /dev/mem device.\n");
		exit (1);
    } else {
    	int mmu_fd;
    	upper_memory = (u8 *)mmap(0, CACHE_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, mem_fd, 0x02000000);
#ifdef MMUHACK
        mmu_fd = open("/dev/mmuhack", O_RDWR);
        if(mmu_fd < 0) {
            system("/sbin/insmod -f mmuhack.o");
            mmu_fd = open("/dev/mmuhack", O_RDWR);
        }
        if(mmu_fd >= 0) close(mmu_fd);
#endif
    }

    screen = (u16 *)mmap(0, CACHE_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, mem_fd, 0x03101000);
    work_frame = (u16 *)malloc(512 * 288 * 2);
#else
	/* Initialize SDL */
	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK | SDL_INIT_TIMER) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit (1);
	}
	atexit (Terminate);

	SDL_ShowCursor(SDL_DISABLE);

	/* Set 320x240 16-bits video mode */
	screen_surface = SDL_SetVideoMode (320, 240, 16, SDL_SWSURFACE );
	if (screen_surface == NULL) {
		fprintf (stderr, "Couldn't set 320x240x16 video mode: %s\n", SDL_GetError ());
		exit (2);
	}
    screen = screen_surface->pixels;

	work_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, 512, 288, 16, 0xf800, 0x07e0, 0x001f, 0x0000);
	work_frame = work_surface->pixels;

	/* Check and open joystick device */
	if (SDL_NumJoysticks() > 0) {
		SDL_JoystickOpen(0);
	}

	SDL_ShowCursor(SDL_DISABLE);
#endif

	option_sound_enable = 1;
	option_samplerate = 1;
	option_controller = INPUT_PLAYER1;
	option_speedlimit = 0;
	option_frameskip = 0;
	option_autoframeskip = 1;

	sound_volume(50);

	{
        int i;
    	memset(resize_table, 0, sizeof(resize_table));
    	for(i=2;i<320;i+=5)resize_table[i] = 1;
    }

	if(argc == 2) {
        char tmp[256];
        char *p = strrchr(argv[1], '/');
        if(p == NULL)
            strcpy(tmp, argv[1]);
        else
            strcpy(tmp, p + 1);
        p = strrchr(tmp, '.');
        if(p != NULL) *p = 0;
        if(strlen(tmp) < 15) strcpy(game_name, tmp);
        printf("Loading %s game...\n", game_name);
    }

	snap_no = -1;
#if (CPS_VERSION == 2)
	cps2_main();
#else
	cps1_main();
#endif
}


static TICKER get_ticks(void)
{
#ifdef GP2X
    struct timeval tv;
    gettimeofday(&tv, 0);
    return (TICKER)tv.tv_sec * TICKS_PER_SEC + tv.tv_usec;
#else
    return SDL_GetTicks();
#endif
}

/*--------------------------------------------------------
	ի죭ૹëתѢ
--------------------------------------------------------*/

void autoframeskip_reset(void)
{
	frameskip = option_autoframeskip ? 0 : option_frameskip;
	frameskipadjust = 0;
	frameskip_counter = FRAMESKIP_LEVELS - 1;

	rendered_frames_since_last_fps = 0;
	frames_since_last_fps = 0;

	game_speed_percent = 100;
	frames_per_second = FPS;

	frames_displayed = 0;

	skip_nextframe = 0;
	next_frame_base = get_ticks() + TICKS_PER_FRAME;
}


/*--------------------------------------------------------
	ի죭ૹë׫ƣ֫
--------------------------------------------------------*/
#if 0
u8 skip_this_frame(void)
{
	return skip_nextframe ? 1 : skiptable[frameskip][frameskip_counter];
}
#endif

/*--------------------------------------------------------
	
--------------------------------------------------------*/

void update_screen(void)
{
//	u8 skipped_it = skip_this_frame(); //skiptable[frameskip][frameskip_counter];
	u8 skipped_it = skiptable[frameskip][frameskip_counter];
#ifndef GP2X
    SDL_Event event;
    int button = -1;
/*
	if (!skipped_it)
	{
		if (option_showfps) show_fps();
		show_battery_warning();
		ui_show_popup(1);
	}
	else ui_show_popup(0);

	if (warming_up)
	{
		video_wait_vsync();
		last_skipcount0_time = SDL_GetTicks() - FRAMESKIP_LEVELS * TICKS_PER_FRAME;
		warming_up = 0;
	}

	if (frameskip_counter == 0)
		this_frame_base = last_skipcount0_time + FRAMESKIP_LEVELS * TICKS_PER_FRAME;

	frames_displayed++;
	frames_since_last_fps++;
*/
    while(SDL_PollEvent(&event)) {
        switch(event.type) {
            case SDL_QUIT:
                Loop = LOOP_EXIT;
                break;
            case SDL_KEYDOWN:
            case SDL_KEYUP:
                button = -1;
                switch(event.key.keysym.sym) {
                    case SDLK_ESCAPE: Loop = LOOP_EXIT; break;
                    case SDLK_1: button = P1_START; break;
                    case SDLK_2: button = P2_START; break;
                    case SDLK_5: button = P1_COIN; break;
                    case 'a': button = P1_BUTTON1; break;
                    case 's': button = P1_BUTTON2; break;
                    case 'd': button = P1_BUTTON3; break;
                    case 'z': button = P1_BUTTON4; break;
                    case 'x': button = P1_BUTTON5; break;
                    case 'c': button = P1_BUTTON6; break;
                    case SDLK_LEFT: button = P1_LEFT; break;
                    case SDLK_RIGHT: button = P1_RIGHT; break;
                    case SDLK_UP: button = P1_UP; break;
                    case SDLK_DOWN: button = P1_DOWN; break;
                }
                if(button != -1)
                    input_flag[button] = (event.type == SDL_KEYDOWN);
                break;
        }
    }
#endif

    if(option_autoframeskip) {
        skipped_it = 1;
        while(get_ticks() < next_frame_base) {
#ifdef WIN32
            SDL_Delay(1);
#endif
            skipped_it = 0;
        }
        //if(skipped_it)printf("skip it...\n");
    }

#ifdef GP2X
    if((option_sound_enable) && (option_sound_volume))
        qsound_update();
#endif

	if (!skipped_it)
	{
        draw_screen();

        {
            register u16 *src = work_frame + BUF_WIDTH * 16 + 64;
            register u16 *dst = screen + 320 * 8;
            register int x, y;
            for(y = 224; y; --y) {
                u16 *line = src;
                u8 *table = resize_table;

                for(x = 320; x; --x) {
                    if(*table++) {
                        int c1 = *line++;
                        int c2 = *line++;
                        *dst++ = COLORMIX(c1, c2);
                    } else
                        *dst++ = *line++;
                }
                memset(src, 0, 384 * 2);
                src += BUF_WIDTH;
            }
        }

#ifndef GP2X
//		SDL_BlitSurface(work_frame, &rect, screen, &dest);
		SDL_Flip(screen_surface);
#endif

		frameskip = option_frameskip;

#if 0
		rendered_frames_since_last_fps++;

		if (frameskip_counter == 0)
		{
			float seconds_elapsed = (float)(curr - last_skipcount0_time) * (1.0 / (float)TICKS_PER_SEC);
			float frames_per_sec = (float)frames_since_last_fps / seconds_elapsed;

			game_speed_percent = (int)(100.0 * frames_per_sec / (float)FPS + 0.5);
			frames_per_second = (int)((float)rendered_frames_since_last_fps / seconds_elapsed + 0.5);

			last_skipcount0_time = curr;
			frames_since_last_fps = 0;
			rendered_frames_since_last_fps = 0;
		}
#endif
	}
	
	if(msg_count) {
        if(!--msg_count) draw_text(NULL);
    }

	frameskip_counter = (frameskip_counter + 1) % FRAMESKIP_LEVELS;
	next_frame_base += TICKS_PER_FRAME;
}


/*--------------------------------------------------------
	FPS
--------------------------------------------------------*/

static void show_fps(void)
{
	int sx;
	char buf[32];

	sprintf(buf, "%s%2d %3d%% %2dfps",
		option_autoframeskip ? "auto" : "fskp",
		frameskip,
		game_speed_percent,
		frames_per_second);

/*
	sx = SCR_WIDTH - (strlen(buf) << 3);
	small_font_print(sx, 0, buf, 1);
*/
}


/*--------------------------------------------------------
	Ыëƫ꣭ͱ
--------------------------------------------------------*/

static void show_battery_warning(void)
{
/*
	if (!scePowerIsBatteryCharging())
	{
		int bat = scePowerGetBatteryLifePercent();

		if (bat < 10)
		{
			static u32 counter = 0;

			counter++;
			if ((counter % 120) < 80)
			{
				char warning[128];

				boxfill_alpha(0, 254, SCR_WIDTH-1, SCR_HEIGHT-1, COLOR_BLACK, 12);
				sprintf(warning, "Warning: Battery is low (%d%%). Please charge battery!", bat);
				uifont_print_center(256, UI_COLOR(UI_PAL_WARNING), warning);
				screen_is_dirty = 2;
			}
		}
	}
*/
}


void msg_screen_init(const char *text)
{
    int i;
    for(i = 320 * 8; i; --i)
        screen[i] = 0x1f;
    memset(screen + 320 * 8, 0, 320 * 232 * 2);
    msg_line = 0;
    msg_printf(text);
}

void msg_screen_clear()
{
    msg_line = -1;
#ifndef GP2X
    SDL_FillRect(screen_surface, NULL, 0);
    SDL_Flip(screen_surface);
#else
    memset(screen, 0, 320 * 240 * 2);
#endif
}

static void draw_text(const char *text)
{
	int x = 0, y;
	u16 *dst = screen;
    const char *p = text;

	if(msg_line == -1) {
        y = 232;
        msg_count = 60*3;
    } else {
        y = msg_line * 8;
        if(msg_line == 30) {
            memcpy(dst + 320 * 8, dst + 320 * 16, 320 * 224 * 2);
            y -= 8;
        } else {
            ++msg_line;
        }
    }

    dst += y * 320;
    if(y != 0)
        memset(dst, 0, 320 * 8 * 2);
    if(p == NULL) return;

    while(*p) {
        if(*p == 0x20) {
            x += 8;
        } else if(*p > 0x20) {
            int w, h;
            u8 *font = (u8 *)&font_data[(*p - 0x20) * 8];
            u16 *put = dst + x;
            
            for(h = 8; h; --h) {
                u8 c = *font++;
                for(w = 8; w; --w, ++put) {
                    if(c & 0x80) *put = 0xffff;
                    c <<= 1;
                }
                put += 320 - 8;
            }
            x += 8;
        }
        ++p;
        if(x > 312) {
            if(msg_line != -1)
                draw_text(p);
            break;
        }
    }
}

void msg_printf(const char *text, ...)
{
	va_list arg;
	char buf[512];

	va_start(arg, text);
	vsprintf(buf, text, arg);
	va_end(arg);
	
	if(msg_line == 0) {	
    	printf("Message : %s\n", buf);
    } else {
        printf("%s", buf);
    }
    
    draw_text(buf);
	
#ifndef GP2X
    SDL_Flip(screen_surface);
#endif
}

void button_wait()
{
#ifdef GP2X
	u32 buttons;
    if(joy_fd < 0) return;

    while(1) {
        usleep(0);
        if((read(joy_fd, &buttons, 4) == 4) && buttons)
            return;
    }
#else
    SDL_Event event;

    while(1) {
        SDL_Delay(1);
        if(SDL_PollEvent(&event)) {
            switch(event.type) {
                case SDL_QUIT:
                    Loop = LOOP_EXIT;
                    return;
                case SDL_KEYDOWN:
                case SDL_KEYUP:
                case SDL_JOYBUTTONDOWN:
                case SDL_JOYBUTTONUP:
                    return;
            }
        }
    }
#endif
}


/*--------------------------------------------------------
	٤ܫ飭ë
--------------------------------------------------------*/

void fatalerror(const char *text, ...)
{
	va_list arg;

	va_start(arg, text);
	vsprintf(fatal_error_message, text, arg);
	va_end(arg);

	fatal_error = 1;
	Loop = LOOP_BROWSER;
}


/*--------------------------------------------------------
	٤ܫ飭ë
--------------------------------------------------------*/

void show_fatal_error(void)
{
	if (fatal_error)
	{
		int sx, sy, ex, ey;
//		int width = uifont_get_string_width(fatal_error_message);
		int update = 1;
		msg_printf("ERROR: %s\n", fatal_error_message);
		button_wait();
		Loop = LOOP_EXIT;
/*
		sx = (SCR_WIDTH - width) >> 1;
		sy = (SCR_HEIGHT - FONTSIZE) >> 1;
		ex = sx + width;
		ey = sy + (FONTSIZE - 1);

		load_background();

		while (Loop != LOOP_EXIT)
		{
			if (update)
			{
				show_background();
				small_icon(6, 3, UI_COLOR(UI_PAL_TITLE), ICON_SYSTEM);
				uifont_print(32, 5, UI_COLOR(UI_PAL_TITLE), "Fatal error");
				draw_dialog(sx - FONTSIZE/2, sy - FONTSIZE/2, ex + FONTSIZE/2, ey + FONTSIZE/2);
				uifont_print_center(sy, UI_COLOR(UI_PAL_SELECT), fatal_error_message);

				update = draw_battery_status(1);
				video_flip_screen(1);
			}
			else
			{
				update = draw_battery_status(0);
				video_wait_vsync();
			}

			pad_update();

			if (pad_pressed(PSP_CTRL_ANY))
				break;
		}

		pad_wait_clear();
*/
		fatal_error = 0;
	}
}


/*------------------------------------------------------
	꣭󫷫ë
------------------------------------------------------*/

void save_snapshot(void)
{
	char path[MAX_PATH];

#if (CPS_VERSION == 2)
	sound_mute(1);
	cache_sleep(1);
#else
	sound_mute(0);
#endif
/*
	if (snap_no == -1)
	{
		FILE *fp;

		snap_no = 1;

		while (1)
		{
			sprintf(path, "%ssnap/%s_%02d.bmp", launchDir, game_name, snap_no);
			if ((fp = fopen(path, "rb")) == NULL) break;
			fclose(fp);
			snap_no++;
		}
	}

	sprintf(path, "%ssnap/%s_%02d.bmp", launchDir, game_name, snap_no);
	save_bmp(path);
	ui_popup("Snapshot saved as \"%s_%02d.bmp\".", game_name, snap_no++);
*/
#if (CPS_VERSION == 2)
	cache_sleep(0);
	sound_mute(0);
#else
	cps1_sound_mute(0);
#endif
}
